#  Copyright (c) 2023. ISTMO Center S.A.  All Rights Reserved
#

import datetime
import os
import time
from typing import Any, Dict

import pandas as pd

from utils.df_utils import add_df_totals

timeseries_options: Dict = {
    'index_col': 'last_activity_dt',
    'rule': 'D',
    'cols': None,
    'vals': None,
    'totals': False,
    'aggfunc': 'count'
}


def convert_value_pairs(row):
    if row['value_type'] == "datetime":
        converted_value = datetime.datetime.strptime(row['value'], '%Y-%m-%d %H:%M:%S')
    elif row['value_type'] == 'str':
        converted_value = row['value']
    else:
        converted_value = eval(f"{row['value_type']}({row['value']})")
    return converted_value


def get_timeseries_params(df_params) -> Any:
    index_col = None
    ts_params = df_params
    if ts_params.get("datetime_index", False):
        index_col = ts_params.get('index_col', None)
        pop_cols = ['datetime_index', 'index_col']
        for p in pop_cols:
            ts_params.pop(p, None)
    return index_col, ts_params


def format_fields(df, format_options):
    for fld_name, fld_type in format_options.items():
        if fld_name in df.columns:
            df[fld_name] = df[fld_name].values.astype(fld_type)
    return df


def fillna_fields(df, fill_options):
    for fld_name, fill_value in fill_options.items():
        if fld_name in df.columns:
            df[fld_name].fillna(fill_value, inplace=True)
    return df


def cast_cols_as_categories(df, threshold=100):
    for col in df.columns:
        if df[col].dtype == object and len(df[col].unique()) < threshold:
            df[col] = df[col].astype(pd.CategoricalDtype())
    return df


def load_as_timeseries(df, **options):
    index_col = options.get("index_col", None)
    if index_col is not None and df.index.name != index_col:
        if index_col in df.columns:
            df.reset_index(inplace=True)
            df.set_index(index_col, inplace=True)
    rule = options.get("rule", "D")
    index = options.get("index", df.index)
    cols = options.get("cols", None)
    vals = options.get("vals", None)
    totals = options.get("totals", False)
    agg_func = options.get("agg_func", 'count')
    df = df.pivot_table(index=index, columns=cols, values=vals, aggfunc=agg_func).fillna(0)
    df = df.resample(rule=rule).sum()
    df.sort_index(inplace=True)
    if totals:
        df = add_df_totals(df)
    return df


def fix_fields(df, fields_to_fix, field_type):
    field_attributes = {
        'str': {'default_value': '', 'dtype': str},
        'int': {'default_value': 0, 'dtype': int},
        'date': {'default_value': pd.NaT, 'dtype': 'datetime64[ns]'},
        'datetime': {'default_value': pd.NaT, 'dtype': 'datetime64[ns]'}
    }

    if field_type not in field_attributes:
        raise ValueError("Invalid field type: {}".format(field_type))

    attr = field_attributes[field_type]
    fields = [field for field in fields_to_fix if field in df.columns]
    df[fields] = df[fields].fillna(attr['default_value']).astype(attr['dtype'])


class DataWrapper:
    def __init__(self, dataclass, date_field, data_path, parquet_filename, start_date, end_date,
                 verbose=False, load_params=None, reverse_order=False, overwrite=False, max_age_minutes=1440):
        self.dataclass = dataclass
        self.date_field = date_field
        self.data_path = self.ensure_forward_slash(data_path)
        self.parquet_filename = parquet_filename
        self.verbose = verbose
        self.load_params = load_params or {}
        self.reverse_order = reverse_order
        self.overwrite = overwrite
        self.max_age_minutes = max_age_minutes

        self.start_date = self.convert_to_date(start_date)
        self.end_date = self.convert_to_date(end_date)

    @staticmethod
    def convert_to_date(date):
        return datetime.datetime.strptime(date, '%Y-%m-%d').date() if isinstance(date, str) else date

    @staticmethod
    def ensure_forward_slash(path):
        return path if path.endswith('/') else path + '/'

    @staticmethod
    def ensure_directory_exists(path):
        os.makedirs(path, exist_ok=True)

    def generate_date_range(self):
        step = -1 if self.reverse_order else 1
        start, end = (self.end_date, self.start_date) if self.reverse_order else (self.start_date, self.end_date)
        current_date = start
        while current_date != end + datetime.timedelta(days=step):
            yield current_date
            current_date += datetime.timedelta(days=step)

    def process(self):
        for current_date in self.generate_date_range():
            self.process_date(current_date)

    def is_file_older_than(self, file_path, current_date):
        """Checks if the file at file_path is older than max_age_minutes or if overwrite is True."""
        today = datetime.date.today()

        if self.overwrite or self.max_age_minutes == 0 or current_date == today:
            return True
        if not os.path.exists(file_path):
            return True  # If the file does not exist, treat as "old" to allow generation.
        file_age_seconds = time.time() - os.path.getmtime(file_path)
        if self.verbose:
            print(f"File {file_path} is {round((file_age_seconds / 60), 0)} minutes old")

        return file_age_seconds / 60 > self.max_age_minutes

    def process_date(self, date):
        folder = f'{self.data_path}{date.year}/{date.month:02d}/{date.day:02d}/'
        self.ensure_directory_exists(folder)
        full_parquet_filename = os.path.join(folder, self.parquet_filename)

        if self.verbose:
            print(f"Processing {full_parquet_filename}...")

        if not self.is_file_older_than(full_parquet_filename, date):
            if self.verbose:
                print("File exists and conditions for regeneration are not met. Skipping.")
            return

        # Assuming dataclass has a method to load data for a specific date and save to parquet
        data_object = self.dataclass(live=True)
        date_filter_params = {
            f'{self.date_field}__year': date.year,
            f'{self.date_field}__month': date.month,
            f'{self.date_field}__day': date.day
        }
        df = data_object.load(**self.load_params, **date_filter_params)

        if df.empty:
            if self.verbose:
                print("No data found for the specified date.")
            return

        data_object.save_to_parquet(df, parquet_full_path=full_parquet_filename)
        if self.verbose:
            print(f"Data saved to {full_parquet_filename}")

# Usage:
# wrapper = DataWrapper(dataclass=YourDataClass, date_field="created_at", data_path="/path/to/data", parquet_filename="data.parquet", start_date="2022-01-01", end_date="2022-12-31")
# wrapper.process()



#  Copyright (c) 2023. ISTMO Center S.A.  All Rights Reserved
#
#
# coding: utf-8
import sys
import django

from django.core.cache import cache
from django.db.models import Field

if sys.version_info >= (3,):
    from django.utils.encoding import force_str as force_text
else:
    from django.utils.encoding import force_text


def is_values_queryset(qs):
    if django.VERSION < (1, 9):  # pragma: no cover
        return isinstance(qs, django.db.models.query.ValuesQuerySet)
    else:
        try:
            return qs._iterable_class == django.db.models.query.ValuesIterable
        except:
            return False


def object_to_dict(obj, fields: list = None):
    """
        Convert obj to a dictionary
        Parameters
        ----------
        obj: obj to an item of QuerySet
        fieldnames: reserved fields, default to all fields
    """
    if not fields:
        obj.__dict__.pop('_state')
        return obj.__dict__
    return {field: obj.__dict__.get(field) for field in fields}


def get_model_name(model):
    """
    Returns the name of the model
    """
    return model._meta.model_name


def replace_from_choices(choices):
    def inner(values):
        return [choices.get(v, v) for v in values]

    return inner


def get_base_cache_key(model):
    return 'pandas_%s_%s_%%s_rendering' % (
        model._meta.app_label, get_model_name(model))


def get_cache_key(obj):
    return get_base_cache_key(obj._meta.model) % obj.pk


def invalidate(obj):
    cache.delete(get_cache_key(obj))


def invalidate_signal_handler(sender, **kwargs):
    invalidate(kwargs['instance'])


def replace_pk(model):
    base_cache_key = get_base_cache_key(model)

    def get_cache_key_from_pk(pk):
        return None if pk is None else base_cache_key % str(pk)

    def inner(pk_series):
        pk_series = pk_series.astype(object).where(pk_series.notnull(), None)
        cache_keys = pk_series.apply(
            get_cache_key_from_pk, convert_dtype=False)
        unique_cache_keys = list(filter(None, cache_keys.unique()))

        if not unique_cache_keys:
            return pk_series

        out_dict = cache.get_many(unique_cache_keys)

        if len(out_dict) < len(unique_cache_keys):
            out_dict = dict([(base_cache_key % obj.pk, force_text(obj))
                             for obj in model.objects.filter(
                    pk__in=list(filter(None, pk_series.unique())))])
            cache.set_many(out_dict)

        return list(map(out_dict.get, cache_keys))

    return inner


def build_update_functions(fieldnames, fields):
    for fieldname, field in zip(fieldnames, fields):
        if not isinstance(field, Field):
            yield fieldname, None
        else:
            if field and field.choices:
                choices = dict([(k, force_text(v))
                                for k, v in field.flatchoices])
                yield fieldname, replace_from_choices(choices)

            elif field and field.get_internal_type() == 'ForeignKey':
                yield fieldname, replace_pk(get_related_model(field))


def update_with_verbose(df, fieldnames, fields):
    for fieldname, function in build_update_functions(fieldnames, fields):
        if function is not None:
            df[fieldname] = function(df[fieldname])


def get_related_model(field):
    """Gets the related model from a related field"""
    model = None

    if hasattr(field, 'related_model') and field.related_model:  # pragma: no cover
        model = field.related_model
    # Django<1.8 doesn't have the related_model API, so we need to use rel,
    # which was removed in Django 2.0
    elif hasattr(field, 'rel') and field.rel:  # pragma: no cover
        model = field.rel.to

    return model



#  Copyright (c) 2023. ISTMO Center S.A.  All Rights Reserved
#

import calendar
import datetime
import sys
import time
from typing import Union

import dateutil
import numpy as np
import pandas as pd

from dateutil.relativedelta import relativedelta
from utils.log_utils import Logger

logger = Logger(log_dir='./logs/', logger_name=__name__, log_file=f'{__name__}.log')


def convert_to_datetime(value):
    """
    This function attempts to convert a value to datetime format.
    The function supports multiple datetime string formats including 'YYYY-MM-DD', 'YYYY-MM-DD HH:MM:SS' and 'YYYY-MM-DDTHH:MM:SSZ'.
    """
    formats = ['%Y-%m-%d', '%Y-%m-%d %H:%M:%S', '%Y-%m-%dT%H:%M:%SZ']

    for fmt in formats:
        try:
            return datetime.datetime.strptime(value, fmt)
        except ValueError:
            continue
    raise ValueError(f'{value} is not in a supported datetime format')


def get_month_day_range(date) -> tuple[datetime.date, datetime.date]:
    """
    For a date 'date' returns the start and end date for the month of 'date'.

    Month with 31 days:
    date = datetime.date(2011, 7, 27)
    get_month_day_range(date)
    (datetime.date(2011, 7, 1), datetime.date(2011, 7, 31))

    Month with 28 days:
    date = datetime.date(2011, 2, 15)
    get_month_day_range(date)
    (datetime.date(2011, 2, 1), datetime.date(2011, 2, 28))
    """
    if isinstance(date, str):
        date = datetime.datetime.strptime(date, "%Y-%m-%d")
    # date = convert_to_datetime(date)
    first_day: datetime.date = date.replace(day=1)
    last_day: datetime.date = date.replace(day=calendar.monthrange(date.year, date.month)[1])
    return first_day, last_day


def get_date_range(year, month):
    start_date = f"{year}-{month:02d}-01"
    end_date = pd.Period(start_date, freq='M').end_time.strftime('%Y-%m-%d')
    return [start_date, end_date]


def get_quarter(date: datetime) -> datetime:
    """
        Returns the quarter number for a given date.
    """
    return (date.month - 1) // 3 + 1


def get_first_day_of_the_quarter(date: datetime) -> datetime:
    """
    This function returns the first day of the quarter for a given date.

    :param date: A datetime object representing a date
    :type date: datetime.datetime
    :return: A datetime object representing the first day of the quarter
    :rtype: datetime.datetime
    """
    if isinstance(date, str):
        date = datetime.datetime.strptime(date, "%Y-%m-%d")
    return datetime.datetime(date.year, 3 * ((date.month - 1) // 3) + 1, 1)


def get_last_day_of_the_quarter(date: datetime) -> datetime:
    """
        This function returns the last day of the quarter for a given date.

        :param date: A datetime object representing a date
        :type date: datetime.datetime
        :return: A datetime object representing the last day of the quarter
        :rtype: datetime.datetime
    """
    if isinstance(date, str):
        date = datetime.datetime.strptime(date, "%Y-%m-%d")
    quarter = get_quarter(date)
    return datetime.datetime(date.year + 3 * quarter // 12, 3 * quarter % 12 + 1, 1) + datetime.timedelta(days=-1)


def get_quarter_date_range(year, quarter):
    quarter_start_month = (quarter - 1) * 3 + 1
    quarter_start = datetime.datetime(year, quarter_start_month, 1)
    quarter_end = quarter_start + datetime.timedelta(days=90)
    return quarter_start, quarter_end


def get_quarter_timerange(date):
    """
    Returns the start and end dates of a quarter for a given year and quarter number.
    """
    if isinstance(date, str):
        date = datetime.datetime.strptime(date, "%Y-%m-%d")
    start = get_day_starttime(get_first_day_of_the_quarter(date))
    end = get_day_endtime(get_last_day_of_the_quarter(date))
    return start, end


def get_day_timerange(date):
    """
        Returns the start and end times of a given date.
    """
    if isinstance(date, str):
        date = datetime.datetime.strptime(date, "%Y-%m-%d")
    start = datetime.datetime.combine(date, datetime.time.min)
    end = datetime.datetime.combine(date, datetime.time.max)
    return start, end


def get_day_starttime(date):
    if isinstance(date, str):
        date = datetime.datetime.strptime(date, "%Y-%m-%d")
    return datetime.datetime.combine(date, datetime.time.min)


def get_day_endtime(date):
    if isinstance(date, str):
        date = datetime.datetime.strptime(date, "%Y-%m-%d")
    return datetime.datetime.combine(date, datetime.time.max)


def remove_tz_from_datefield(df, date_fields):
    for date_field in date_fields:
        df[date_field] = pd.to_datetime(df[date_field], errors='coerce')
        df[date_field].fillna(pd.to_datetime(datetime.datetime.today(), format='%Y-%m-%d'), inplace=True)
        df[date_field] = df[date_field].apply(
            lambda a: datetime.datetime.strftime(a, "%Y-%m-%d %H:%M:%S"))
        df[date_field] = pd.to_datetime(df[date_field])
    return df


def calc_week_range(date):
    if isinstance(date, str):
        date = datetime.datetime.strptime(date, "%Y-%m-%d")
    current_weekday = date.weekday()
    monday = date - datetime.timedelta(days=current_weekday)
    sunday = monday + datetime.timedelta(days=6)
    return monday, sunday


def get_current_month_period():
    start, end = get_month_day_range(datetime.datetime.today())
    start = get_day_starttime(start)
    end = get_day_endtime(end)
    return start, end


def get_year_timerange(year):
    year_start = datetime.datetime.combine(
        datetime.datetime(year, 1, 1), datetime.time.min)
    year_end = datetime.datetime.combine(datetime.datetime(year, 12, 31), datetime.time.max)
    return year_start, year_end


def get_period_as_dict(start, end):
    return {
        'start': start,
        'end': end,
    }


def is_datetime_column(df, col):
    is_dt_column = False
    unique_values = df[col].unique()
    num_unique = len(unique_values)
    converted = pd.to_datetime(unique_values, errors='coerce', infer_datetime_format=True)
    num_converted = (~converted.isna()).sum()
    if num_converted / num_unique >= 0.5:
        is_dt_column = True
    return is_dt_column


def cast_col_as_datetime(df):
    for col in df.columns:
        if df[col].dtype == object:
            is_dt = is_datetime_column(df, col)
            if is_dt:
                date_format = '%Y-%m-%d %H:%M:%S'
                infer_dt = True
                if len(df[df[col].astype(str).str.contains(":") > 0]) == 0:
                    date_format = '%Y-%m-%d'
                    infer_dt = False
                df[col] = pd.to_datetime(df[col], format=date_format, errors='coerce', infer_datetime_format=infer_dt)
                df[col].fillna(method='ffill', inplace=True)
    return df


class BusinessDays:

    def __init__(self, holiday_list):
        """
            Initialize a BusinessDays object with a given holiday list.
        """
        self.HOLIDAY_LIST = holiday_list
        bd_holidays = [day for year in self.HOLIDAY_LIST for day in self.HOLIDAY_LIST[year]]
        self.bd_cal = np.busdaycalendar(holidays=bd_holidays, weekmask='1111100')

    def get_business_days_count(self, begin_date, end_date):
        """
        Calculate the number of business days between two dates.
        """
        # Validate input dates and convert to datetime objects
        try:
            begin_date = datetime.datetime.strptime(begin_date, '%Y-%m-%d')
            end_date = datetime.datetime.strptime(end_date, '%Y-%m-%d')
        except ValueError:
            raise ValueError('Dates should be strings in the format YYYY-MM-DD')

        # Validate that years are in the holiday list
        years = [str(year) for year in range(begin_date.year, end_date.year + 1)]
        if not all(year in self.HOLIDAY_LIST for year in years):
            raise ValueError('Not all years in date range are in the holiday list')

        return np.busday_count(begin_date.strftime('%Y-%m-%d'), end_date.strftime('%Y-%m-%d'), busdaycal=self.bd_cal)

    def calc_business_days_from_df(self, df, begin_date_col, end_date_col, result_col='business_days'):
        # Validate column names
        if not all(col in df.columns for col in [begin_date_col, end_date_col]):
            raise ValueError('Column names not found in DataFrame')
        if df.empty:
            # DataFrame is empty. Cannot perform operation.
            return
        # Ensure dates are in string format
        df[begin_date_col] = pd.to_datetime(df[begin_date_col]).dt.strftime('%Y-%m-%d')
        df[end_date_col] = pd.to_datetime(df[end_date_col]).dt.strftime('%Y-%m-%d')

        # Vectorize np.busday_count function
        v_func = np.vectorize(np.busday_count, excluded=['busdaycal'])

        # Apply vectorized function to DataFrame
        df[result_col] = v_func(df[begin_date_col], df[end_date_col], busdaycal=self.bd_cal)

    def add_business_days(self, start_date, n_days):
        """
        Add n_days business days to start_date.
        """
        # Validate input dates and convert to datetime objects
        try:
            start_date = datetime.datetime.strptime(start_date, '%Y-%m-%d')
        except ValueError:
            raise ValueError('Date should be a string in the format YYYY-MM-DD')

        # Validate that year is in the holiday list
        if str(start_date.year) not in self.HOLIDAY_LIST:
            logger.warning(f'Year {start_date.year} is not in the holiday list')

        return np.busday_offset(start_date.strftime('%Y-%m-%d'), n_days, roll='forward', busdaycal=self.bd_cal)

    def calc_sla_end_date(self, df, start_date_col, n_days_col, result_col='sla_end_date'):
        """
        Calculate SLA end date based on start date and number of SLA days.
        """
        # Validate column names
        if not all(col in df.columns for col in [start_date_col, n_days_col]):
            raise ValueError('Column names not found in DataFrame')
        if df.empty:
            # DataFrame is empty. Cannot perform operation.
            return
        # Ensure dates are in string format
        df[start_date_col] = pd.to_datetime(df[start_date_col]).dt.strftime('%Y-%m-%d')

        # Vectorize np.busday_offset function
        v_func = np.vectorize(self.add_business_days, excluded=['busdaycal'])

        # Apply vectorized function to DataFrame
        df[result_col] = v_func(df[start_date_col], df[n_days_col])


def get_today():
    return get_today_timerange()


def get_yesterday():
    return get_period_as_dict(get_day_starttime(datetime.datetime.today() - datetime.timedelta(days=1)),
                              get_day_endtime(datetime.datetime.today() - datetime.timedelta(days=1)))


def get_current_month():
    start, end = get_current_month_period()
    return get_period_as_dict(start, end)


def get_previous_month():
    start, end = get_month_day_range(datetime.datetime.today() - relativedelta(months=1))
    start = get_day_starttime(start)
    end = get_day_endtime(end)
    return get_period_as_dict(start, end)


def get_current_year():
    start, end = get_year_timerange(datetime.datetime.today().year)
    return get_period_as_dict(start, end)


def get_today_timerange():
    return get_period_as_dict(get_day_starttime(datetime.datetime.today()),
                              get_day_endtime(datetime.datetime.today()))


def get_current_quarter():
    start, end = get_quarter_timerange(datetime.datetime.today())
    return get_period_as_dict(start, end)


def get_quarter_period(year, quarter):
    start, end = get_quarter_date_range(year, quarter)
    return get_period_as_dict(start, end)


def get_year_period(year):
    start, end = get_year_timerange(year)
    return get_period_as_dict(start, end)


def get_week_range():
    start, end = calc_week_range(datetime.datetime.today())
    return get_period_as_dict(start, end)


def get_last_week():
    start, end = calc_week_range(datetime.datetime.today() - datetime.timedelta(days=7))
    return get_period_as_dict(start, end)


def count_down(msg, maximum):
    for i in range(maximum, 0, -1):
        pad_str = ' ' * len('%d' % 1)
        sys.stdout.write('%s for the next %d seconds %s\r' %
                         (msg, i, pad_str), )
        sys.stdout.flush()
        time.sleep(1)


def load_period(**kwargs):
    start = kwargs.get('start')
    end = kwargs.get('end')
    return get_period_as_dict(start, end)


def load_timeframe(timeframe_func, **kwargs):
    return load_period(**timeframe_func(**kwargs))


def load_current_week():
    return load_timeframe(get_week_range)


def load_current_month():
    return load_timeframe(get_current_month)


def load_current_quarter():
    return load_timeframe(get_current_quarter)



#  Copyright (c) 2023. ISTMO Center S.A.  All Rights Reserved
#
import datetime
from typing import Dict, Union
import json
import pandas as pd
from django.db import models
# This is the configuration file for the dataframe_helper module.

# conversion_map is a dictionary that maps the field types to their corresponding data type conversion functions.
# Each entry in the dictionary is a pair of a field type (as a string) and a callable function that performs the
# conversion. This mapping is used to convert the values in a pandas DataFrame to the appropriate data types based on
# the field type.

conversion_map: Dict[str, callable] = {
    'CharField': lambda x: x.astype(str),
    'TextField': lambda x: x.astype(str),
    'IntegerField': lambda x: pd.to_numeric(x, errors='coerce'),
    'AutoField': lambda x: pd.to_numeric(x, errors='coerce'),
    'BigIntegerField': lambda x: pd.to_numeric(x, errors='coerce'),
    'SmallIntegerField': lambda x: pd.to_numeric(x, errors='coerce'),
    'PositiveIntegerField': lambda x: pd.to_numeric(x, errors='coerce'),
    'PositiveSmallIntegerField': lambda x: pd.to_numeric(x, errors='coerce'),
    'FloatField': lambda x: pd.to_numeric(x, errors='coerce'),
    'DecimalField': lambda x: pd.to_numeric(x, errors='coerce'),
    'BooleanField': lambda x: x.astype(bool),
    'NullBooleanField': lambda x: x.astype(bool),
    'DateTimeField': lambda x: pd.to_datetime(x, errors='coerce'),
    # 'DateField': lambda x: pd.to_datetime(x, errors='coerce'),
    'DateField': lambda x: pd.to_datetime(x, errors='coerce').dt.date,
    # 'DateField': lambda x: datetime.datetime.combine(x, datetime.datetime.min.time()),
    'TimeField': lambda x: pd.to_datetime(x, errors='coerce').dt.time,
    'DurationField': lambda x: pd.to_timedelta(x, errors='coerce'),
    # for JSONField, assuming JSON objects are represented as string in df
    'JSONField': lambda x: x.apply(json.loads),
    'ArrayField': lambda x: x.apply(eval),
    'UUIDField': lambda x: x.astype(str),
}

# default_kwargs is a dictionary that provides default values for various optional parameters.
# These default values are used when the corresponding parameters are not provided by the user.

default_kwargs: Dict[str, Union[str, bool, models.Model,int, Dict, None]] = {
    'connection_name': None,
    'table': None,
    'model': None,
    'field_map': {},
    'params': None,
    'legacy_filters': False,
    'live': False,
    'use_exclude': False,
    'debug': False,
    'n_records': 0,
    'dt_field': None,
    'use_dask': False,
    'verbose': False,
}
# dataframe_params is a dictionary that provides configuration options for creating a pandas DataFrame.
# These options include parameters for specifying the columns, index column, and other options for DataFrame creation.

dataframe_params: Dict[str, Union[tuple, str, bool, int, None]] = {
    "fieldnames": (),
    "index_col": None,
    "coerce_float": False,
    "verbose": True,
    "datetime_index": False,
    "column_names": None
}
# dataframe_options is a dictionary that provides additional options for modifying a pandas DataFrame.
# These options include parameters for handling duplicate values, sorting, grouping, and other DataFrame operations.

dataframe_options: Dict[str, Union[bool, str, int, None]] = {
    "debug": False,  # Whether to print debug information
    "duplicate_expr": None,  # Expression for identifying duplicate values
    "duplicate_keep": 'last',  # How to handle duplicate values ('first', 'last', or False)
    "sort_field": None,  # Field to use for sorting the DataFrame
    "group_by_expr": None,  # Expression for grouping the DataFrame
    "group_expr": None  # Expression for aggregating functions to the grouped DataFrame
}

parquet_default_params: Dict[str, Union[bool, str, int, None]] = {
    "use_parquet": False,  # Whether to use parquet files for storing data
    "parquet_storage_path": None,  # Path to the folder where parquet files are stored
    "parquet_filename": None,  # Name of the parquet file
    "parquet_max_age_minutes": None,  # Maximum age of the parquet file in minutes
}





#  Copyright (c) 2024. ISTMO Center S.A.  All Rights Reserved
#  IBIS is a registered trademark
#
import datetime
import os
from pathlib import Path
from typing import Optional, Dict, Union, Any, List

import dask.dataframe as dd
from pydantic import BaseModel, model_validator, DirectoryPath, FilePath

from ..file_utils import FilePathGenerator
from ..log_utils import Logger
from .config import dataframe_params, dataframe_options
from .sql_model_builder import SqlModelBuilder

logger = Logger(log_dir='./logs/', logger_name=__name__, log_file=f'{__name__}.log')

LOOKUP_SEP = "__"


class ConnectionConfig(BaseModel):
    live: bool = False
    connection_name: str = None
    table: str = None
    model: Any = None

    @model_validator(mode='after')
    def check_model(self):
        if self.connection_name is None:
            raise ValueError('Connection name must be specified')

        if self.live is False:
            if self.model is None:
                raise ValueError('Model must be specified')
            self.table = self.model._meta.db_table
        else:
            # if live is True, then connection_name and table must be specified
            if self.table is None:
                raise ValueError('Table name must be specified')
            self.model = SqlModelBuilder(connection_name=self.connection_name, table=self.table).build_model()

        return self


class QueryConfig(BaseModel):
    use_exclude: bool = False
    n_records: int = 0
    dt_field: Optional[str] = None
    use_dask: bool = False
    as_dask: bool = False

    @model_validator(mode='after')
    def check_n_records(self):
        if self.n_records < 0:
            raise ValueError('Number of records must be non-negative')


class ParamsConfig(BaseModel):
    field_map: Optional[Dict] = None
    legacy_filters: bool = False
    fixed_filters: Dict[str, Union[str, bool, int, float]] = {}
    filters: Dict[str, Union[str, Dict, bool, int, float]] = {}
    df_params: Dict[str, Union[tuple, str, bool, None]] = {}
    df_options: Dict[str, Union[bool, str, None]] = {}
    params: Dict[str, Union[str, bool, int, float, List[Union[str, int, bool, float]]]] = {}

    @model_validator(mode='after')
    def check_params(self):
        if self.params is not None:
            self.parse_params(self.params)
        return self

    def parse_params(self, params, use_parquet=False):
        df_params, df_options, filters = {}, {}, {}
        for k, v in params.items():
            if k in dataframe_params:
                df_params.update({k: v})
            elif k in dataframe_options:
                df_options.update({k: v})
            else:
                filters.update({k: v})
        if not self.fixed_filters:
            self.fixed_filters = filters
        self.filters = {**self.fixed_filters, **filters}
        self.df_params = {**self.df_params, **df_params}
        self.df_options = {**self.df_options, **df_options}
        if self.legacy_filters and not use_parquet:
            self.convert_legacy_filters()

    def convert_legacy_filters(self):
        if not self.legacy_filters or not self.field_map or not self.filters:
            return

        # create a reverse map of the field_map
        reverse_map = {v: k for k, v in self.field_map.items()}

        new_filters = {}
        for filter_field, value in self.filters.items():
            # split the filter_field if LOOKUP_SEP exists
            parts = filter_field.split(LOOKUP_SEP, 1)

            # replace each part with its legacy equivalent if it exists
            new_parts = [reverse_map.get(part, part) for part in parts]

            # join the parts back together and add to the new filters
            new_filter_field = LOOKUP_SEP.join(new_parts)
            new_filters[new_filter_field] = value

        self.filters = new_filters
        return self


class ParquetOptions(BaseModel):
    df_as_dask: bool = False
    use_parquet: bool = False
    save_parquet: bool = False
    load_parquet: bool = False
    parquet_filename: Optional[str] = None
    parquet_storage_path: Optional[DirectoryPath] = None
    parquet_full_path: Optional[FilePath] = None
    parquet_folder_list: Optional[List[str]] = None
    parquet_size_bytes: int = 0
    parquet_max_age_minutes: int = 0
    parquet_is_recent: bool = False
    parquet_start_date: str = None
    parquet_end_date: str = None

    @model_validator(mode='after')
    def check_parquet_params(self):
        if self.use_parquet is False:
            return self
        self.save_parquet = False
        self.load_parquet = False
        if self.parquet_storage_path is None:
            raise ValueError('Parquet storage path must be specified')
        if not os.path.exists(self.parquet_storage_path):
            raise ValueError('Parquet storage path does not exist')
        if self.parquet_filename is not None:
            self.parquet_full_path = self.ensure_file_extension(
                filepath=self.parquet_storage_path / self.parquet_filename, extension='parquet')
            self.parquet_is_recent = self.is_file_recent()
            self.load_parquet = self.parquet_is_recent and self.parquet_full_path.exists()
            self.save_parquet = not self.parquet_is_recent

        if self.parquet_start_date is not None:
            if self.parquet_end_date is None:
                raise ValueError('Parquet end date must be specified if start date is provided')

            start_date = datetime.datetime.strptime(self.parquet_start_date, '%Y-%m-%d')
            end_date = datetime.datetime.strptime(self.parquet_end_date, '%Y-%m-%d')
            if end_date < start_date:
                raise ValueError('Parquet end date must be greater than start date')

            # saving to parquet is disabled when start and end dates are provided, since we will be
            # just loading the parquet files
            self.save_parquet = False
            self.parquet_folder_list = FilePathGenerator(str(self.parquet_storage_path)).generate_file_paths(start_date,
                                                                                                             end_date)
            self.parquet_size_bytes = self.get_parquet_size_bytes()
            self.load_parquet = all(
                [os.path.exists(folder) for folder in self.parquet_folder_list]) and self.parquet_size_bytes > 0
        elif self.parquet_end_date is not None:
            raise ValueError('Parquet start date must be specified if end date is provided')

        return self

    def is_file_recent(self):
        if not os.path.exists(self.parquet_full_path):
            return False
        if self.parquet_max_age_minutes == 0:
            return True
        file_time = datetime.datetime.fromtimestamp(os.path.getmtime(self.parquet_full_path))
        if datetime.datetime.now() - file_time > datetime.timedelta(minutes=self.parquet_max_age_minutes):
            # logger.info(f"File {self.parquet_full_path} is older than {self.parquet_max_age_minutes} minutes")
            return False
        return True

    def get_parquet_size_bytes(self):
        total_size = 0
        for folder in self.parquet_folder_list:
            for root, dirs, files in os.walk(folder):
                for file in files:
                    if file.endswith('.parquet'):
                        total_size += os.path.getsize(os.path.join(root, file))
        return total_size

    @staticmethod
    def ensure_file_extension(filepath: Path, extension: str) -> Path:
        if not filepath.suffix == f".{extension}":
            return filepath.with_suffix(f".{extension}")
        return filepath

    @staticmethod
    def apply_filters_dask(df, filters):
        dt_operators = ['date', 'time']
        date_operators = ['year', 'month', 'day', 'hour', 'minute', 'second', 'week_day']
        comparison_operators = [
            'gte',
            'lte',
            'gt',
            'lt',
            'exact',
            'in',
            'range',
            'contains',
            'icontains',
            'startswith',
            'endswith',
            'isnull'
        ]

        operation_map = {
            'exact': lambda col, val: col == val,
            'gt': lambda col, val: col > val,
            'gte': lambda col, val: col >= val,
            'lt': lambda col, val: col < val,
            'lte': lambda col, val: col <= val,
            'in': lambda col, val: col.isin(val),
            'range': lambda col, val: (col >= val[0]) & (col <= val[1]),
            'contains': lambda col, val: col.str.contains(val, regex=True),
            'icontains': lambda col, val: col.str.contains(val, case=False),
            'startswith': lambda col, val: col.str.startswith(val),
            'endswith': lambda col, val: col.str.endswith(val),
            'isnull': lambda col, val: col.isnull() if val else col.notnull(),
        }

        def get_temp_col(dask_df, field_name, casting):
            if casting in dt_operators + date_operators:
                # Ensure datetime conversion for date/time operations
                temp_col = dd.to_datetime(dask_df[field_name]) if casting in dt_operators else dask_df[field_name]
                if casting in date_operators + dt_operators:
                    temp_col = getattr(temp_col.dt, casting)
            else:
                temp_col = df[field_name]
            return temp_col

        for key, value in filters.items():
            parts = key.split('__')
            field_name = parts[0]
            casting = None
            operation = 'exact'

            if len(parts) == 3:
                # Adjust logic based on the parts
                _, casting, operation = parts
            elif len(parts) == 2:
                # Could be either a casting or an operation
                if parts[1] in comparison_operators:
                    operation = parts[1]
                elif parts[1] in dt_operators + date_operators:
                    casting = parts[1]
            # field_name, casting, operation = (parts + [None, 'exact'])[:3]  # Defaults added to match expected length
            temp_col = get_temp_col(df, field_name, casting)

            if operation in operation_map:
                condition = operation_map[operation](temp_col, value)
                df = df[condition]
            else:
                raise ValueError(f"Unsupported operation: {operation}")

        return df

        # dt_operators = ['date', 'time']
        # date_operators = ['year', 'month', 'day', 'hour', 'minute', 'second']
        # comparison_operators = [
        #     'gte',
        #     'lte',
        #     'gt',
        #     'lt',
        #     'exact',
        #     'in',
        #     'range',
        #     'contains',
        #     'icontains',
        #     'startswith',
        #     'endswith',
        #     'isnull'
        # ]
        # for key, value in filters.items():
        #     parts = key.split('__')
        #
        #     # Initialize defaults
        #     field_name = parts[0]
        #     casting = None
        #     operation = 'exact'  # Default operation if not explicitly provided
        #
        #     if len(parts) == 3:
        #         # Adjust logic based on the parts
        #         _, casting, operation = parts
        #     elif len(parts) == 2:
        #         # Could be either a casting or an operation
        #         if parts[1] in comparison_operators:
        #             operation = parts[1]
        #         elif parts[1] in dt_operators + date_operators:
        #             casting = parts[1]
        #
        #     # Prepare the column for operation
        #     if casting in dt_operators + date_operators:
        #         # Temporarily convert to datetime for operation, if not already
        #         temp_col = dd.to_datetime(df[field_name]) if casting in dt_operators else df[field_name]
        #     else:
        #         temp_col = df[field_name]
        #
        #     if casting in date_operators + dt_operators:
        #         temp_col = getattr(temp_col.dt, casting)
        #
        #     if operation == 'exact':
        #         condition = temp_col == value
        #     elif operation == 'gt':
        #         condition = temp_col > value
        #     elif operation == 'gte':
        #         condition = temp_col >= value
        #     elif operation == 'lt':
        #         condition = temp_col < value
        #     elif operation == 'lte':
        #         condition = temp_col <= value
        #     elif operation == 'in':
        #         condition = temp_col.isin(value)
        #     elif operation == 'range':
        #         condition = (temp_col >= value[0]) & (temp_col <= value[1])
        #     elif operation == 'contains':
        #         condition = temp_col.str.contains(value, regex=True)
        #     elif operation == 'icontains':
        #         condition = temp_col.str.contains(value, case=False)
        #     elif operation == 'startswith':
        #         condition = temp_col.str.startswith(value)
        #     elif operation == 'endswith':
        #         condition = temp_col.str.endswith(value)
        #     elif operation == 'isnull':
        #         condition = temp_col.isnull() if value else temp_col.notnull()
        #     df = df[condition]
        #
        # return df



#  Copyright (c) 2024. ISTMO Center S.A.  All Rights Reserved
#  IBIS is a registered trademark
#
import dask.dataframe as dd
from django.db.models import Q

from ..log_utils import Logger
from .config import *
from .io import read_frame
from .models import ParquetOptions, QueryConfig, ConnectionConfig, ParamsConfig

logger = Logger(log_dir='./logs/', logger_name=__name__, log_file=f'{__name__}.log')


class DfHelper:
    debug: bool = False
    verbose: bool = False
    live: bool = False
    parquet_options: ParquetOptions = None
    query_config: QueryConfig = None
    connection_config: ConnectionConfig = None

    def __init__(self, **kwargs):
        self.debug = kwargs.pop('debug', False)
        self.verbose = kwargs.pop('verbose', False)
        self.live = kwargs.get('live', False)
        kwargs['live'] = self.live
        self.parquet_options = ParquetOptions(**kwargs)
        if self.parquet_options.use_parquet is False:
            self.connection_config = ConnectionConfig(**kwargs)
        else:
            self.connection_config = None
        self.query_config = QueryConfig(**kwargs)
        self.params_config = ParamsConfig(**kwargs)

    def load(self, **options):
        return self._load(**options)

    def _load(self, **options: Dict[str, Union[str, bool, Dict, None]]) -> pd.DataFrame:
        # if options is not None:
        #self.params_config.filters = self.params_config.fixed_filters
        self.params_config.parse_params(options, use_parquet=self.parquet_options.use_parquet)
        if self.parquet_options.use_parquet is True:
            result = self.load_from_parquet()
        else:

            result = self._load_from_db()

        return result

    def load_from_parquet(self) -> pd.DataFrame:
        if self.debug:
            logger.info(f'Loading from parquet file...{self.parquet_options.parquet_full_path}')
            if self.verbose:
                print(f'Loading from parquet file...{self.parquet_options.parquet_full_path}')
        try:
            if self.parquet_options.parquet_folder_list:
                df = dd.read_parquet(self.parquet_options.parquet_folder_list)
            else:
                df = dd.read_parquet(self.parquet_options.parquet_full_path)
            df = self.parquet_options.apply_filters_dask(df, self.params_config.filters)
            if self.parquet_options.df_as_dask:
                return df
            else:
                return df.compute()
        except FileNotFoundError:
            if self.debug:
                logger.info(f'Parquet file {self.parquet_options.parquet_full_path} does not exist')
                if self.verbose:
                    print(f'Parquet file {self.parquet_options.parquet_full_path} does not exist')
            if self.parquet_options.df_as_dask:
                return dd.from_pandas(pd.DataFrame(), npartitions=1)
            return pd.DataFrame()

    def _load_from_db(self) -> pd.DataFrame:
        """
        Load data from the database based on the filters provided. If no filters are given, it loads only the first
        n_records.
            n_records: Number of records to load if no filters are provided.
            :return: A DataFrame loaded with data from the database.
        """
        if self.connection_config.model is None:
            if self.debug:
                logger.critical('Model must be specified')
                if self.verbose:
                    print('Model must be specified')
            raise ValueError('Model must be specified')
        df = self._build_and_load()
        if df is not None:
            df = self._process_loaded_data(df)
        return df

    def _build_and_load(self) -> pd.DataFrame:
        query = self.connection_config.model.objects.using(self.connection_config.connection_name)
        if self.debug:
            logger.debug(query.query)
            logger.debug(self.params_config.filters)
        if not self.params_config.filters:
            # IMPORTANT: if no filters are provided show only the first n_records
            # this is to prevent loading the entire table by mistake
            n_records = 100
            if self.debug:
                logger.info(f'No filters provided, showing first %s records, {n_records}')
                if self.verbose:
                    print(f'No filters provided, showing first {n_records} records')
            queryset = query.all()[:n_records]
        else:
            q_objects = self.build_q_objects(self.params_config.filters, self.query_config.use_exclude)
            queryset = query.filter(q_objects)[
                       :self.query_config.n_records] if self.query_config.n_records > 0 else query.filter(q_objects)

        if queryset is not None:
            if self.debug:
                logger.debug(queryset.query)
                if self.verbose:
                    print(queryset.query)

            df = read_frame(queryset, **self.params_config.df_params)
            # df = self.batched_read_frame(qs, 100, **self.dataframe_params)
            if self.debug:
                logger.debug(df.head())
                if self.verbose:
                    print(df.head())
            return df
        else:
            return pd.DataFrame()

    @staticmethod
    def build_q_objects(filters: dict, use_exclude: bool):
        q_objects = Q()
        for key, value in filters.items():
            if not use_exclude:
                q_objects.add(Q(**{key: value}), Q.AND)
            else:
                q_objects.add(~Q(**{key: value}), Q.AND)
        return q_objects

    def _process_loaded_data(self, df: pd.DataFrame) -> pd.DataFrame:
        df = self._convert_columns(df)
        if self.params_config.field_map:
            if self.debug:
                logger.info(f'Renaming columns...{[col for col in self.params_config.field_map.keys()]}')
                if self.verbose:
                    print(f'Renaming columns...{[col for col in self.params_config.field_map.keys()]}')
            set_to_keep1 = set(self.params_config.field_map.keys())
            set_to_keep2 = set(self.params_config.df_params.get('column_names', []))
            columns_to_keep = list(set_to_keep1.union(set_to_keep2))
            cols_to_drop = [col for col in df.columns if col not in columns_to_keep]
            df.drop(columns=cols_to_drop, inplace=True)
            df.rename(columns=self.params_config.field_map, inplace=True)
        return df

    def _convert_columns(self, df: pd.DataFrame) -> pd.DataFrame:
        """
                Convert the data types of columns in the DataFrame based on the field type in the Django model.

                :param df: DataFrame whose columns' data types are to be converted.
                :return: DataFrame with converted column data types.
        """
        if self.debug:
            logger.info(f'Converting columns:{[col for col in df.columns]}')
        model_fields = self.connection_config.model._meta.get_fields()
        for field in model_fields:
            field_name = field.name
            field_type = type(field).__name__
            if field_name in df.columns:
                if self.debug:
                    logger.debug(f"Found column {field_name} of type {field_type}")
                    if self.verbose:
                        print(f"Found column {field_name} of type {field_type}")
                if field_type in list(conversion_map.keys()):
                    try:
                        df[field_name] = conversion_map[field_type](df[field_name])
                        if self.debug:
                            logger.info(f"Converted column {field_name} of type {field_type}")
                            if self.verbose:
                                print(f"Converted column {field_name} of type {field_type}")
                    except Exception as e:
                        if self.debug:
                            logger.info(f"Error converting column {field_name} of type {field_type}")
                            logger.error(e)
                            if self.verbose:
                                print(f"Error converting column {field_name} of type {field_type}")
                else:
                    if self.debug:
                        logger.error(f"Field type {field_type} not found in conversion_map")
                        if self.verbose:
                            print(f"Field type {field_type} not found in conversion_map")
            else:
                if self.debug:
                    logger.error(f"Column {field_name} not found in df.columns")
                    if self.verbose:
                        print(f"Column {field_name} not found in df.columns")
        return df

    def save_to_parquet(self, df: pd.DataFrame, parquet_full_path, engine: str = 'auto') -> None:
        """IMPORTANT: This routine must be explicitly called to save the
        dataframe to a parquet file This is because sometimes the first generated df is not what we want to save but
        saving after cleaning data and after other merge operations
        """
        if self.debug:
            logger.info(f'Saving to parquet file...{parquet_full_path}')
            if self.verbose:
                print(f'Saving to parquet file...{parquet_full_path}')
        if not df.empty:
            try:
                df.to_parquet(parquet_full_path, engine=engine)
            except Exception as e:
                logger.error(f'Error while saving DataFrame to parquet: {e}')
                if self.verbose:
                    print(f'Error while saving DataFrame to parquet: {e}')
        else:
            logger.warning(f'Attempted to save an empty DataFrame to {parquet_full_path}')
            if self.verbose:
                print(f'Attempted to save an empty DataFrame to {parquet_full_path}')



#  Copyright (c) 2023. ISTMO Center S.A.  All Rights Reserved
#

import configparser
import datetime
import glob
import os
import time
import zipfile

import openpyxl
import pandas as pd


class IniFile(configparser.ConfigParser):
    def __init__(self, config_file, defaults=None, overwrite_defaults=False):
        super().__init__()
        self.config_file = config_file
        self.defaults = defaults
        self.overwrite_defaults = overwrite_defaults
        self.go()

    def go(self):
        if not os.path.exists(self.config_file):
            self.save_config()
            self.overwrite_defaults = True

        if self.defaults is not None and self.overwrite_defaults:
            for section, content in self.defaults.items():
                self.add_section(section)
                for key in content.keys():
                    self.set(section, key, str(content[key]))
            self.save_config()
        self.read(self.config_file)
        return self

    def get_config(self, section: str, option: str) -> str:
        return self.get(section, option)

    def write_config(self, section: str, option: str, value: str) -> None:
        self.set(section, option, str(value))
        self.save_config()

    def save_config(self) -> None:
        with open(self.config_file, 'w') as configfile:
            self.write(configfile)


#
# def ensure_file_extension(filename, extension):
#     basename, ext = os.path.splitext(filename)
#     if ext != extension:
#         new_filename = f"{basename}.{extension}"
#         # shutil.move(filename, new_filename)
#         return new_filename
#     return filename


# Alternative Version
# class IniFile:
#     def __init__(self, config_file, defaults=None, overwrite_defaults=False):
#         self.config_file = config_file
#         self.defaults = defaults
#         self.overwrite_defaults = overwrite_defaults
#         self.config = configparser.ConfigParser()
#         self.create_file()
#         self.add_defaults()
#         self.read_file()
#
#     def create_file(self):
#         if not os.path.exists(self.config_file):
#             with open(self.config_file, 'w') as f:
#                 pass
#
#     def add_defaults(self):
#         if self.defaults is not None and self.overwrite_defaults:
#             for section, content in self.defaults.items():
#                 self.config.add_section(section)
#                 for key in content.keys():
#                     self.config.set(section, key, str(content[key]))
#             self.save_config()
#
#     def read_file(self):
#         self.config.read(self.config_file)
#
#     def get_config(self, section: str, option: str) -> str:
#         return self.config.get(section, option)
#
#     def write_config(self, section: str, option: str, value: str) -> None:
#         self.config.set(section, option, str(value))
#
#     def save_config(self) -> None:
#         with open(self.config_file, 'w') as configfile:
#             self.config.write(configfile)

def check_make_dir(dirpath):
    if not os.path.exists(dirpath):
        os.makedirs(dirpath, exist_ok=True)
    return os.path.join(dirpath, '')


def ensure_directory_exists(dir_path):
    os.makedirs(dir_path, exist_ok=True)
    return dir_path


def ensure_forward_slash(dir_path):
    if not dir_path.endswith('/'):
        return f"{dir_path}/"
    return dir_path


def compress_files_to_zip(file_list, zip_file_name):
    with zipfile.ZipFile(zip_file_name, 'w') as zip_file:
        for file in file_list:
            zip_file.write(file)
    return zip_file_name


def adjust_worksheet_columns(worksheet, df):
    for col_idx, col in enumerate(df.columns):
        series = df[col]
        if isinstance(series, pd.Series):
            # Find the maximum length of the column name and the values in the column
            max_len = max(
                series.astype(str).map(len).max(),
                len(str(series.name))
            ) + 1  # Adding a little extra space
            # Set the width of the column in the worksheet
            worksheet.column_dimensions[openpyxl.utils.get_column_letter(col_idx + 1)].width = max_len
            # worksheet.set_column(col_idx, col_idx, max_len)


def read_most_recent_parquet(dir_path, starts_with=None) -> pd.DataFrame:
    """Read the most recent Parquet file in a specified directory."""

    # List all Parquet files in the directory that start with the specified prefix
    if starts_with is None:
        files = [f for f in os.listdir(dir_path) if f.endswith('.parquet')]
    else:
        files = [f for f in os.listdir(dir_path) if f.endswith('.parquet') and f.startswith(starts_with)]

    # Ensure there is at least one file
    if not files:
        raise ValueError("No Parquet files found in directory matching the criteria")

    # Find the most recent file
    most_recent_file = max(files, key=lambda x: os.path.getmtime(os.path.join(dir_path, x)))

    # Define the file path
    file_path = os.path.join(dir_path, most_recent_file)

    # Load the DataFrame from the Parquet file
    df = pd.read_parquet(file_path)

    return df


def ensure_file_extension(filename, extension):
    if not filename.lower().endswith(f'.{extension}'):
        return f"{filename}.{extension}"
    else:
        return filename


def check_file_exists(filename):
    return os.path.exists(filename)


def handle_save_options(**kwargs):
    save = kwargs.pop('save', False)
    filename = kwargs.pop('filename', None)
    extension = kwargs.pop('extension', None)
    stub = kwargs.pop('stub', None)

    if filename is not None:
        filename = ensure_file_extension(filename, extension)
        save = True

    if save and filename is None:
        filename = f"{stub}_{datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S')}.{extension}"

    return save, filename


def check_file_is_recent(filename, age=0, unit='minutes'):
    if not check_file_exists(filename):
        return True
    if age == 0:
        return True
    if unit == 'years':
        age = age * 31536000
    elif unit == 'months':
        age = age * 2628000
    elif unit == 'days':
        age = age * 86400
    elif unit == 'hours':
        age = age * 3600
    elif unit == 'minutes':
        age = age * 60
    else:
        raise ValueError("Invalid unit specified")

    if os.path.getmtime(filename) > time.time() - age:
        return True
    else:
        return False


def prune_aged_files(dir_path, age_days):
    """Prune files that are older than a certain number of days."""

    # Get the current time
    now = time.time()

    # Walk through the directory
    for root, dirs, files in os.walk(dir_path):
        for file in files:
            file_path = os.path.join(root, file)

            # If the file is older than the specified number of days, delete it
            if os.path.getmtime(file_path) < now - age_days * 86400:
                os.remove(file_path)
                print(f"Deleted {file_path}")


def load_newest_file(dir_path, extension='xlsx'):
    files = os.listdir(dir_path)
    files = [os.path.join(dir_path, f) for f in files if f.endswith(f'.{extension}')]
    files.sort(key=lambda x: os.path.getmtime(x))
    recent_file = files[-1]
    return recent_file


def fix_sheet_name(sheet_name):
    invalid_chars = [":", "\\", "/", "?", "*", "[", "]"]
    for char in invalid_chars:
        sheet_name = sheet_name.replace(char, '-')
    sheet_name = sheet_name[:30]
    return sheet_name


class FilePathGenerator:
    # Example Usage:
    # generator = FilePathGenerator(base_path='/your/base/path')
    # file_paths = generator.generate_file_paths('2022-01-01', '2022-03-31')
    # print(file_paths)
    def __init__(self, base_path=''):
        self.base_path = base_path.rstrip('/')

    @staticmethod
    def _get_day_file_path(dir_path, day):
        return f"{dir_path}/{str(day).zfill(2)}/*.parquet"

    @staticmethod
    def _get_month_file_path(dir_path):
        return f"{dir_path}/*/*.parquet"

    def generate_file_paths(self, start_date, end_date):
        start_date = self._convert_to_datetime(start_date)
        end_date = self._convert_to_datetime(end_date)

        file_paths = []
        curr_date = start_date

        while curr_date <= end_date:
            year, month = curr_date.year, curr_date.month
            dir_path = f"{self.base_path}/{year}/{str(month).zfill(2)}"

            if os.path.exists(dir_path):
                file_paths.extend(self._get_files_for_month(curr_date, start_date, end_date, dir_path))

            curr_date = self._increment_month(curr_date)

        return file_paths

    @staticmethod
    def _convert_to_datetime(date):
        if isinstance(date, str):
            return datetime.datetime.strptime(date, '%Y-%m-%d')
        return date

    def _get_files_for_month(self, curr_date, start_date, end_date, dir_path):
        files = []
        if curr_date.year == start_date.year and curr_date.month == start_date.month:
            if curr_date.year == end_date.year and curr_date.month == end_date.month:
                files.extend(self._get_files_for_range(dir_path, start_date.day, end_date.day))
            else:
                start_day = start_date.day if start_date.day > 1 else 1
                files.extend(self._get_files_for_range(dir_path, start_day, 31))  # Max days in a month
        elif curr_date.year == end_date.year and curr_date.month == end_date.month:
            files.extend(self._get_files_for_range(dir_path, 1, end_date.day))
        else:
            month_file_path = self._get_month_file_path(dir_path)
            if glob.glob(month_file_path):
                files.append(month_file_path)
        return files

    def _get_files_for_range(self, dir_path, start_day, end_day):
        files = []
        for day in range(start_day, end_day + 1):
            day_file_path = self._get_day_file_path(dir_path, day)
            if glob.glob(day_file_path):
                files.append(day_file_path)
        return files

    @staticmethod
    def _increment_month(curr_date):
        if curr_date.month == 12:
            return datetime.datetime(curr_date.year + 1, 1, 1)
        else:
            return datetime.datetime(curr_date.year, curr_date.month + 1, 1)





#  Copyright (c) 2023. ISTMO Center S.A.  All Rights Reserved
#
from .log_utils import Logger

logger = Logger(log_dir='./logs/', logger_name=__name__, log_file=f'{__name__}.log')


def load_grouped_activity(df, **kwargs):
    debug = kwargs.pop('debug', False)
    group_by_expr = kwargs.pop('group_by_expr', None)
    if group_by_expr is None:
        raise ValueError('group_by_expr must be specified')
    group_expr = kwargs.pop('group_expr', 'count')
    if group_by_expr is not None:
        if debug:
            logger.info("Grouping by: {}".format(group_by_expr))
        df = df.groupby(by=group_by_expr).size().reset_index(name=group_expr)
    return df


def summarise_data(df, **opts):
    summary_columns = opts.get("summary_column", None)
    if summary_columns is None:
        raise ValueError('summary_column must be specified')
    value_columns = opts.get("values_column", None)
    if value_columns is None:
        raise ValueError('values_column must be specified')
    resample_rule = opts.get("rule", "D")
    agg_func = opts.get("agg_func", 'count')
    df = df.pivot_table(index=df.index, columns=summary_columns, values=value_columns,
                        aggfunc=agg_func).fillna(0)
    df = df.resample(resample_rule).sum()
    return df


def summarize_and_resample_data(df, summary_columns, value_columns, **opts):
    if summary_columns is None:
        raise ValueError('summary_column must be specified')
    if value_columns is None:
        raise ValueError('values_column must be specified')

    resample_rule = opts.get("rule", "D")
    agg_func = opts.get("agg_func", 'count')

    return (df.pivot_table(index=df.index, columns=summary_columns, values=value_columns, aggfunc=agg_func)
            .fillna(0)
            .resample(resample_rule)
            .sum())


def load_latest(df, **kwargs):
    kwargs.update({'duplicate_keep': 'last'})
    return eval_duplicate_removal(df, **kwargs)


def load_earliest(df, **kwargs):
    kwargs.update({'duplicate_keep': 'first'})
    return eval_duplicate_removal(df, **kwargs)


def add_df_totals(df):
    df.loc['Total'] = df.sum(numeric_only=True, axis=0)
    df.loc[:, 'Total'] = df.sum(numeric_only=True, axis=1)
    return df


def eval_duplicate_removal(df, **df_options):
    duplicate_expr = df_options.get('duplicate_expr', None)
    debug = df_options.get('debug', False)
    if duplicate_expr is None:
        return df
    if debug:
        df_duplicates = df[df.duplicated(duplicate_expr)]
        print("Duplicate Rows based on columns are:", df_duplicates, sep='\n')
    sort_field = df_options.get('sort_field', None)
    keep_which = df_options.get('duplicate_keep', 'last')
    if sort_field is None:
        df = df.drop_duplicates(duplicate_expr, keep=keep_which)
    else:
        df = df.sort_values(sort_field).drop_duplicates(duplicate_expr, keep=keep_which)
    return df



#  Copyright (c) 2023. ISTMO Center S.A.  All Rights Reserved
#
#

import django

from django.core.paginator import Paginator
import pandas as pd

from .utils import update_with_verbose, get_related_model, object_to_dict, is_values_queryset

FieldDoesNotExist = (
    django.db.models.fields.FieldDoesNotExist
    if django.VERSION < (1, 8)
    else django.core.exceptions.FieldDoesNotExist
)


def to_fields(qs, fieldnames):
    """
    Get fields from a queryset based on the given fieldnames.

    :param qs: The queryset from which to extract fields.
    :param fieldnames: A list of fieldnames to extract.
    :return: An iterator that yields the corresponding fields from the queryset.
    """
    for fieldname in fieldnames:
        model = qs.model
        for fieldname_part in fieldname.split('__'):
            try:
                field = model._meta.get_field(fieldname_part)
            except FieldDoesNotExist:
                try:
                    rels = model._meta.get_all_related_objects_with_model()
                except AttributeError:
                    field = fieldname
                else:
                    for relobj, _ in rels:
                        if relobj.get_accessor_name() == fieldname_part:
                            field = relobj.field
                            model = field.model
                            break
            else:
                model = get_related_model(field)
        yield field


def read_frame(qs, fieldnames=(), index_col=None, coerce_float=False,
               verbose=True, datetime_index=False, column_names=None):
    """
        Read a DataFrame from a queryset or list of objects.

        :param qs: The queryset or list of objects.
        :param fieldnames: The field names to include in the DataFrame. If not provided, all fields will be included.
        :param index_col: The name of the column to use as the index in the DataFrame.
        :param coerce_float: Whether to coerce all values to floats.
        :param verbose: Whether to print verbose output during the process.
        :param datetime_index: Whether to convert the index column to datetime if it contains datetime values.
        :param column_names: The names of the columns in the DataFrame. If not provided, the field names will be used.
        :return: The DataFrame.

        .. note::
            - If `fieldnames` is provided, only the specified fields will be included in the DataFrame.
            - If `index_col` is provided, the specified column will be used as the index in the DataFrame.
            - If `coerce_float` is True, all values will be coerced to floats.
            - If `verbose` is True, verbose output will be printed during the process.
            - If `datetime_index` is True, the index column will be converted to datetime if it contains datetime values.
            - If `column_names` is provided, the specified names will be used as column names in the DataFrame.
    """

    if fieldnames:
        fieldnames = pd.unique(fieldnames)
        if index_col is not None and index_col not in fieldnames:
            # Add it to the field names if not already there
            fieldnames = tuple(fieldnames) + (index_col,)
            if column_names:
                column_names = tuple(column_names) + (index_col,)
        fields = to_fields(qs, fieldnames)
    elif is_values_queryset(qs):
        # if django.VERSION < (1, 9):  # pragma: no cover
        #     annotation_field_names = list(qs.query.annotation_select)
        #
        #     if annotation_field_names is None:
        #         annotation_field_names = []
        #
        #     extra_field_names = qs.extra_names
        #     if extra_field_names is None:
        #         extra_field_names = []
        #
        #     select_field_names = qs.field_names
        #
        # else:  # pragma: no cover
        annotation_field_names = list(qs.query.annotation_select)
        extra_field_names = list(qs.query.extra_select)
        select_field_names = list(qs.query.values_select)

        fieldnames = select_field_names + annotation_field_names + \
                     extra_field_names
        fields = [None if '__' in f else qs.model._meta.get_field(f)
                  for f in select_field_names] + \
                 [None] * (len(annotation_field_names) + len(extra_field_names))

        uniq_fields = set()
        fieldnames, fields = zip(
            *(f for f in zip(fieldnames, fields)
              if f[0] not in uniq_fields and not uniq_fields.add(f[0])))
    else:
        try:
            fields = qs.model._meta.fields
            fieldnames = [f.name for f in fields]
            fieldnames += list(qs.query.annotation_select.keys())
        except:
            pass

    if is_values_queryset(qs):
        recs = list(qs)
    else:
        try:
            recs = list(qs.values_list(*fieldnames))
            # recs = list(qs.values_list(*fieldnames).iterator())
        except:
            if fieldnames:
                recs = [object_to_dict(q, fieldnames) for q in qs]
                # recs = [object_to_dict(q, fieldnames) for q in qs.iterator()]
            else:
                recs = [object_to_dict(q) for q in qs]
                # recs = [object_to_dict(q) for q in qs.iterator()]

    df = pd.DataFrame.from_records(
        recs,
        columns=column_names if column_names else fieldnames,
        coerce_float=coerce_float
    )

    if verbose:
        update_with_verbose(df, fieldnames, fields)

    if index_col is not None:
        df.set_index(index_col, inplace=True)

    if datetime_index:
        df.index = pd.to_datetime(df.index, errors="ignore")
    return df



#  Copyright (c) 2024. ISTMO Center S.A.  All Rights Reserved
#  IBIS is a registered trademark
#
from __future__ import annotations

import sys as _sys
from typing import Any as _Any

from ._df_helper import *

__all__ = ['DfHelper']
